@@ -1,6 +1,15 @@
-"""
-Perform cross validation.
-"""
+# -*- coding: utf-8 -*-
+
+__auther__ ='aeiwz'
+author_email='theerayut_aeiw_123@hotmail.com'
+__copyright__="Copyright 2024, Theerayut"
+
+__license__ = "MIT"
+__maintainer__ = "aeiwz"
+__email__ = "theerayut_aeiw_123@hotmail.com"
+__status__ = "Develop"
+
+
 import collections
 import typing
 import numpy as np
