@@ -1,5 +1,15 @@
+# -*- coding: utf-8 -*-
+
+__auther__ ='aeiwz'
+author_email='theerayut_aeiw_123@hotmail.com'
+__copyright__="Copyright 2024, Theerayut"
+
+__license__ = "MIT"
+__maintainer__ = "aeiwz"
+__email__ = "theerayut_aeiw_123@hotmail.com"
+__status__ = "Develop"
+
 
-#Preprocessing data set to decrease noise
 class Denoise:
     def decrease_noise(spectra, window_length=11, polyorder=2):
         import pandas as pd
