@@ -1,5 +1,14 @@
 # -*- coding: utf-8 -*-
 
+__auther__ ='aeiwz'
+author_email='theerayut_aeiw_123@hotmail.com'
+__copyright__="Copyright 2024, Theerayut"
+
+__license__ = "MIT"
+__maintainer__ = "aeiwz"
+__email__ = "theerayut_aeiw_123@hotmail.com"
+__status__ = "Develop"
+
 
 # Import necessary libraries
 import numpy as np
@@ -30,7 +39,7 @@ from .scaler import Scaler
 
 import os
 
-__auther__ = "aeiwz"
+
 
 
 class opls_da:
@@ -153,10 +162,47 @@ class opls_da:
         
     
     def __init__(self, X, y, features_name=None, n_components=2, scaling_method='pareto', kfold=3, estimator='opls', random_state=42, auto_ncomp=True):
-        
 
+        '''
+        Purpose:
+        Initializes the model, validates the inputs, and preprocesses the data for further analysis.
+
+        Parameters:
+
+            •	X: array-like (Data matrix)
+        The input features for the model (must be a 2D array-like structure, such as pandas.DataFrame or numpy.ndarray).
+            •	y: array-like (Target vector)
+        The target variable or class labels associated with the rows of X (can be a pandas.Series, numpy.ndarray, or list).
+            •	features_name: array-like, optional (default=None)
+        List of feature names corresponding to the columns of X. If not provided, column names will be generated or inferred.
+            •	n_components: int, optional (default=2)
+        The number of components to use for OPLS-DA analysis. Must be a positive integer.
+            •	scaling_method: str, optional (default=‘pareto’)
+        The method used for scaling the data. Options include:
+            •	'pareto': Pareto scaling (power 0.5)
+            •	'mean': Mean-centered data
+            •	'uv': Unit variance scaling (standardization)
+            •	'minmax': Min-max scaling
+            •	kfold: int, optional (default=3)
+        The number of folds for cross-validation.
+            •	estimator: str, optional (default=‘opls’)
+        The estimator method used for modeling. Default is 'opls' (Orthogonal Projections to Latent Structures).
+            •	random_state: int, optional (default=42)
+        The random seed used for reproducibility.
+            •	auto_ncomp: bool, optional (default=True)
+        If True, automatically selects the optimal number of components for the model. If False, the number of components is set manually.
+
+        Raises:
+
+	        •	ValueError: If any input is invalid, such as mismatched dimensions or incorrect data types.
+
+        '''
         self.auto_ncomp = auto_ncomp
         #check X and y must be dataframe or array
+        if X is None or y is None:
+            raise ValueError("Input data X and target y must be provided.")
+        if n_components < 1:
+            raise ValueError("Number of components must be a positive integer.")
         if not isinstance(X, (pd.DataFrame, np.ndarray)):
             raise ValueError('X must be a dataframe or array')
         if not isinstance(y, (pd.Series, np.ndarray, list)):
@@ -214,6 +260,22 @@ class opls_da:
         
     def fit(self):
         
+        '''        
+        Purpose:
+        Fits the OPLS-DA model to the data and computes model performance metrics.
+
+        Parameters:
+        None
+
+        Returns:
+
+            •	Prints a summary of the model fit, including:
+            •	Sample size and class distributions
+            •	Number of features and components
+            •	Scaling method used
+            •	R² and Q² metrics
+
+        '''
         import time
         
         T1 = time.time()
@@ -286,12 +348,14 @@ class opls_da:
         
         duration = T2 - T1
 
+        unique_values = y.unique()
+
         if self.estimator == 'opls':
     
             summary_model_oplsda = f'''
             Comparison of {y.unique()[0]} and {y.unique()[1]}
-            Sample size: {y.unique()[0]}: {y.value_counts()[0]}
-                         {y.unique()[1]}: {y.value_counts()[1]}
+            Sample size: {y.unique()[0]}: {y.value_counts().get(unique_values[0], 0)}
+                         {y.unique()[1]}: {y.value_counts().get(unique_values[1], 0)}
             Number of features: {X.shape[1]}
             Number of components: {n_components}
             Method of scaling: {self.scaling_method}
@@ -304,8 +368,8 @@ class opls_da:
         else:
             summary_model_plsda = f'''
             Comparison of {y.unique()[0]} and {y.unique()[1]}
-            Sample size: {y.unique()[0]}: {y.value_counts()[0]}
-                         {y.unique()[1]}: {y.value_counts()[1]}
+            Sample size: {y.unique()[0]}: {y.value_counts().get(unique_values[0], 0)}
+                         {y.unique()[1]}: {y.value_counts().get(unique_values[1], 0)}
             Number of features: {X.shape[1]}
             Number of components: {n_components}
             Method of scaling: {self.scaling_method}
@@ -483,22 +547,36 @@ class opls_da:
                  xaxis_direction = 'reversed'):
 
         '''
-        Plot VIP score
+        Purpose:
+        Plots the VIP scores for the features, allowing for customization and thresholding.
 
-        Parameters
-        ----------
-        x_range : int, default=9
-            Range of x-axis.
-        threshold : int, default=2
-            Threshold of VIP score.
-        size : int, default=12
-            Size of the marker.
-        width : int, default=1000
-            Width of the figure.
-        height : int, default=500
-            Height of the figure.
-        filter_ : bool, default=False
-            If True, filter VIP score based on threshold.
+        Parameters:
+
+            •	x_range: int, optional (default=9)
+        The range of the x-axis in the plot.
+            •	threshold: int, optional (default=2)
+        The threshold used to categorize VIP scores into “High” or “Low” importance.
+            •	marker_size: int, optional (default=12)
+        The size of the markers in the scatter plot.
+            •	fig_width: int, optional (default=1000)
+        The width of the plot.
+            •	fig_height: int, optional (default=500)
+        The height of the plot.
+            •	filter_: bool, optional (default=False)
+        If True, only features with VIP scores above the threshold are plotted.
+            •	vip_transform: bool, optional (default=True)
+        If True, the VIP scores are transformed to reflect their contribution to the model.
+            •	font_size: int, optional (default=20)
+        The font size for labels.
+            •	title_font_size: int, optional (default=20)
+        The font size for the plot title.
+            •	xaxis_direction: str, optional (default=‘reversed’)
+        The direction of the x-axis (‘reversed’ or ‘normal’).
+
+        Returns:
+
+            •	fig: Plotly figure
+        A Plotly scatter plot visualizing the VIP scores for the features.
 
         '''
 
@@ -592,7 +670,7 @@ class opls_da:
     def plot_oplsda_scores(self,x_ = 't_scores', y_ = 't_ortho', color_ = None, color_dict = None, 
                            symbol_ = None, symbol_dict = None, 
                            fig_height = 900, fig_width = 1300,
-                           marker_size = 35, marker_opacity = 0.7, 
+                           marker_size = 35, marker_opacity = 0.7, marker_label=None, 
                            font_size = 20, title_font_size = 21,
                            legend_name = ['Group', 'Time point'],
                            individual_ellipse = True):
@@ -619,6 +697,22 @@ class opls_da:
         marker_opacity : float, default=0.7
             Opacity of the marker.
 
+        •	Purpose: This function generates an OPLS-DA (Orthogonal Partial Least Squares Discriminant Analysis) scores plot, showing how samples are positioned based on their scores in two principal components (t_scores and t_ortho).
+        •	Parameters:
+            •	x_, y_: The names of the columns in the DataFrame (df_opls_scores) that contain the scores for the x and y axes (default to 't_scores' and 't_ortho').
+            •	color_: An array that assigns a color to each sample (optional).
+            •	color_dict: A dictionary of color mappings for the groups (optional).
+            •	symbol_: An array of symbols for the samples (optional).
+            •	fig_height, fig_width: Dimensions for the plot.
+            •	marker_size, marker_opacity: Control the appearance of the markers.
+            •	legend_name: Custom labels for the legend.
+            •	individual_ellipse: Whether to add ellipses to individual groups (default True).
+        •	Plot Details:
+            •	Uses Plotly (px.scatter) to create an interactive scatter plot.
+            •	Can display confidence ellipses around each group.
+            •	Adds annotations for R²X, R²Y, and Q² statistics.
+            •	The plot is highly customizable (marker size, opacity, labels, colors, etc.).            
+
         '''
         if color_ is not None:
             if len(color_) != len(self.y):
@@ -654,39 +748,24 @@ class opls_da:
 
         #If user not input color_dict then get unique of label and create color_dict
         if color_dict is not None:
-            color_dict = color_dict
+            color_dict_2 = color_dict
         else:
+            
             import plotly.colors as plotly_colour
+
             name_color_set = ['Plotly', 'D3', 'G10', 'T10', 'Alphabet', 'Dark24', 'Light24', 'Set1', 'Pastel1', 
-                        'Dark2', 'Set2', 'Pastel2', 'Set3', 'Antique', 'Safe', 'Bold', 'Pastel', 
-                        'Vivid', 'Prism']
+                                'Dark2', 'Set2', 'Pastel2', 'Set3', 'Antique', 'Safe', 'Bold', 'Pastel', 
+                                'Vivid', 'Prism']
 
             palette = []
             for name in name_color_set:
                 palette += getattr(plotly_colour.qualitative, name) # This is a list of colors
 
+            color_dict = {i: palette[i] for i in range(len(df_opls_scores['Group'].unique()))}
 
-            color_dict = {label: palette[i] for i, label in enumerate(df_opls_scores['Group'].unique())}
-            
-
-        # Create a new color dictionary to match with unique labels
-        color_dict_2 = {}
-        for group in df_opls_scores['Group'].unique():
-            if group in color_dict:
-                color_dict_2[group] = color_dict[group]
-            else:
-                # Assign a default color if the label is not found in the original color_dict
-
-                import plotly.colors as plotly_colour
-                name_color_set = ['Plotly', 'D3', 'G10', 'T10', 'Alphabet', 'Dark24', 'Light24', 
-                                    'Set1', 'Pastel1', 'Dark2', 'Set2', 'Pastel2', 'Set3', 
-                                    'Antique', 'Safe', 'Bold', 'Pastel', 'Vivid', 'Prism']
-
-                palette = []
-                for name in name_color_set:
-                    palette += getattr(plotly_colour.qualitative, name) # This is a list of colors
-
-                color_dict_2[group] = palette[len(color_dict_2) % len(palette)]
+            group_unique = df_opls_scores['Group'].unique()
+            color_dict_2 = {group_unique[i]: list(color_dict.values())[i] for i in range(len(group_unique))}
+        
 
 
         
@@ -694,7 +773,8 @@ class opls_da:
         fig = px.scatter(df_opls_scores, x=x_, y=y_, symbol=symbol_,     
                         symbol_map=symbol_dict,
                         color='Group', 
-                        color_discrete_map=color_dict_2, 
+                        color_discrete_map=color_dict_2,
+                        text=marker_label,
                         title=f'<b>OPLS-DA Scores Plot<b> ({self.scaling_method} scaling)', 
                         height=fig_height, width=fig_width,
                         labels={
@@ -711,6 +791,9 @@ class opls_da:
                             opacity=marker_opacity, 
                             line=dict(width=2, color='DarkSlateGrey')))
 
+        fig.update_traces(textposition='middle center',
+                            textfont_size=marker_size-(0.4*marker_size))
+
 
         fig.add_annotation(dict(font=dict(color="black",size=font_size),
                                 #x=x_loc,
@@ -797,6 +880,16 @@ class opls_da:
         fig_width : int, default=1000
             Width of the figure.
 
+
+	    •	Purpose: This function creates a histogram of permutation scores from a permutation test, commonly used to evaluate model stability and significance.
+	    •	Parameters:
+            •	nbins_: Number of bins in the histogram.
+            •	fig_height, fig_width: Dimensions of the plot.
+            •	font_size, title_font_size: Font size for labels and title.
+            •	Plot Details:
+            •	Plots the permutation scores as a histogram with Plotly.
+            •	Marks the actual model accuracy score with a red dashed line.
+            •	Adds additional annotations for the number of permutations, the accuracy score, and the p-value.
         '''
 
 
@@ -885,6 +978,17 @@ class opls_da:
         color_continuous_scale_ : str, default='jet'
             color_ scale for the plot.
 
+        •	Purpose: This function generates a scatter plot (S-plot), which visualizes the covariance and correlation between the scores of the model.
+	    •	Parameters:
+            •	fig_height, fig_width: Dimensions of the plot.
+            •	range_color_: Range of colors to display.
+            •	color_continuous_scale_: Color scale for the plot.
+            •	marker_size, font_size, title_font_size: Customize marker size and font sizes.
+	    •	Plot Details:
+            •	The plot visualizes the relationship between covariance and correlation for features.
+            •	Uses Plotly’s scatter plot to create an interactive S-plot.
+            •	The axes are customizable, and the plot is set to be visually clean (e.g., axes lines and tick marks).
+
         '''
 
 
@@ -899,7 +1003,7 @@ class opls_da:
                          height=fig_height, width=fig_width,
                          hover_data={'covariance':True, 'correlation':True, 'Features':True})
         
-        fig.update_layout(title='<b>S-plot</b>', xaxis_title='Covariance', yaxis_title='Correlation')
+        fig.update_layout(title='<b>S-plot</b>', xaxis_title='Covariance', yaxis_title='Correlation coefficient')
 
         
         #add line of axis and set color_ to black and line width to 2 pixel
@@ -948,6 +1052,21 @@ class opls_da:
         color_continuous_scale_ : str, default='jet'
             color_ scale for the plot.
 
+
+	    •	Purpose: This function generates a loading plot, typically used in multivariate analysis to visualize the relationship between features and the scores.
+	    •	Parameters:
+            •	fig_height, fig_width: Dimensions of the plot.
+            •	range_color_: Color range to represent the covariance values.
+            •	color_continuous_scale_: Color scale used for continuous color mapping.
+            •	marker_size, font_size, title_font_size: Customize the appearance of the markers and fonts.
+            •	xaxis_direction: Set the direction for the x-axis (e.g., reversed or not).
+            •	xaxis_title: Title for the x-axis.
+	    •	Plot Details:
+            •	The loading plot shows the relationship between features (usually a set of variables or compounds) and the model scores.
+            •	The correlation for each feature is displayed alongside its covariance, with colors representing the covariance values.
+            •	The plot is interactive and customizable (e.g., marker size, color scale, axis settings).
+
+
         '''
 
         s_df_scores_ = self.s_df_scores_
@@ -973,7 +1092,7 @@ class opls_da:
 
         fig.update_traces(marker=dict(size=marker_size))
         fig.update_xaxes(autorange=xaxis_direction)
-        fig.update_layout(title='<b>Loading spectra</b>', xaxis_title=xaxis_title, yaxis_title='Correlation')
+        fig.update_layout(title='<b>Loading spectra</b>', xaxis_title=xaxis_title, yaxis_title='Correlation coefficient')
         fig.update_xaxes(showline=True, linewidth=2, linecolor='black')
         fig.update_yaxes(showline=True, linewidth=2, linecolor='black')
 
@@ -1377,6 +1496,7 @@ class pca:
     def plot_pca_scores(self, pc=['PC1', 'PC2'], 
                         color_=None, color_dict=None, 
                         symbol_=None, symbol_dict=None, 
+                        marker_label=None,
                         fig_height=900, fig_width=1300,
                         marker_size=35, marker_opacity=0.7,
                         font_size=20, title_font_size=21,
@@ -1474,7 +1594,7 @@ class pca:
 
         #If user not input color_dict then get unique of label and create color_dict
         if color_dict is not None:
-            color_dict = color_dict
+            color_dict_2 = color_dict
         else:
             
             import plotly.colors as plotly_colour
@@ -1488,18 +1608,16 @@ class pca:
                 palette += getattr(plotly_colour.qualitative, name) # This is a list of colors
 
             color_dict = {i: palette[i] for i in range(len(df_scores_['Group'].unique()))}
-            
 
-        #new color_dict to match with unique label
-        group_unique = df_scores_['Group'].unique()
-        #change key of color_dict to match with unique label
-        color_dict_2 = {group_unique[i]: list(color_dict.values())[i] for i in range(len(group_unique))}
+            group_unique = df_scores_['Group'].unique()
+            color_dict_2 = {group_unique[i]: list(color_dict.values())[i] for i in range(len(group_unique))}
         
 
         fig = px.scatter(df_scores_, x=pc[0], y=pc[1], color='Group',
                         symbol=symbol_, 
                         color_discrete_map=color_dict_2, 
-                        symbol_map=symbol_dict, 
+                        symbol_map=symbol_dict,
+                        text=marker_label,
                         title=f'<b>PCA Scores Plot<b> {self.scaling_method} scaling', 
                         height=fig_height, width=fig_width,
                         labels={'color': legend_name[0], 'symbol': legend_name[1],
@@ -1513,6 +1631,9 @@ class pca:
                             opacity=marker_opacity, 
                             line=dict(width=2, color='DarkSlateGrey')))
 
+        fig.update_traces(textposition='middle center',
+                            textfont_size=marker_size-(0.4*marker_size))                            
+
 
         fig.add_annotation(dict(font=dict(color="black",size=font_size),
                                 #x=x_loc,
@@ -1622,239 +1743,120 @@ class pca:
 
 
 
-    def plot_pca_trajectory(self, time_, time_order, stat_ = ['mean', 'sem'], pc=['PC1', 'PC2'],
-                            color_dict = None, 
-                            symbol_dict = None, 
+    def plot_pca_trajectory(self, time_, time_order, stat_=['mean', 'sem'], pc=['PC1', 'PC2'],
+                            color_dict=None, symbol_dict=None, 
                             fig_height=900, fig_width=1300,
                             marker_size=35, marker_opacity=0.7,
                             title_font_size=20, font_size=20,
                             legend_name=['Group', 'Time point']):
-
-        '''
-
-        Visualise PCA trajectory
-
-        Parameters
-        ----------
-        time_ : array-like, shape (n_samples,)
-            Time point of samples.
-        time_order : dictionary
-            Order of time point.
-        stat_ : list, default=['mean', 'sem']  
-            Statistic to calculate. First element is mean or median, second element is sem or std.
-        pc : list, default=['PC1', 'PC2']
-            Principle component to plot.
-        color_dict : dictionary, default=None
-            Dictionary of color_ for each group.
-        symbol_dict : dictionary, default=None
-            Dictionary of symbol_ for each time point.
-        fig_height : int, default=900
-            Height of figure.
-        fig_width : int, default=1300
-            Width of figure.
-        marker_size : int, default=35
-            Size of marker.
-        marker_opacity : float, default=0.7
-            Opacity of marker.
-
-        Returns
-        -------
-        fig : plotly.graph_objects.Figure
-            Plotly figure.
-
-        '''
-
-
         from .pca_ellipse import confidence_ellipse
+        import plotly.graph_objects as go
+        import plotly.express as px
+        import pandas as pd
+        import numpy as np
 
-
-        #check time_order must be a dictionary
         if not isinstance(time_order, dict):
-            raise ValueError("time_order must be a dictionary \n Example: time_order = {'Day 1': 0, 'Day 2': 1, 'Day 3': 2}")
+            raise ValueError("`time_order` must be a dictionary.")
+        if time_ is None or len(time_) != len(self.label):
+            raise ValueError("`time_` must match the length of label and not be None.")
+        if stat_[0] not in ['mean', 'median'] or stat_[1] not in ['sem', 'std']:
+            raise ValueError("`stat_` must be ['mean' or 'median', 'sem' or 'std'].")
 
-
-
-
-        #check time are not missing
-        if time_ is None:
-            raise ValueError('time_ must be provided fot time trajectory analysis')
-
-        #check time_ must be pandas series, list or array
-        if not isinstance(time_, (pd.Series, np.ndarray, list)):
-            raise ValueError('time_ must be a series, list or array')
-        if len(time_) != len(self.label):
-            raise ValueError('time_ must have the same number of samples as group')
-
-        
-
-
-        df_scores_ = self.df_scores_
-        r2 = self.df_explained_variance_
-        q2_test = self.q2_test
+        # Copy scores and add time point
+        df_scores_ = self.df_scores_.copy()
         df_scores_['Time point'] = time_
-        
-
-
-
-        #check stat_[0] must be mean or median
-        if stat_[0] not in ['mean', 'median']:
-            raise ValueError('stat_[0] must be mean or median')
-        #check stat_[1] must be sem, std
-        if stat_[1] not in ['sem','std']:
-            raise ValueError('stat_[1] must be sem or std')
 
+        # Aggregate means/medians
         if stat_[0] == 'mean':
             df_scores_point = df_scores_.groupby(['Group', 'Time point']).mean()
-        if stat_[0] == 'median':
+        else:
             df_scores_point = df_scores_.groupby(['Group', 'Time point']).median()
 
+        # Aggregate errors
         if stat_[1] == 'sem':
             err_df = df_scores_.groupby(['Group', 'Time point']).sem()
-        if stat_[1] == 'std':
+        else:
             err_df = df_scores_.groupby(['Group', 'Time point']).std()
 
+        # Map order for sorting
+        df_scores_point['Time order'] = df_scores_point.index.get_level_values('Time point').map(time_order)
+        err_df['Time order'] = err_df.index.get_level_values('Time point').map(time_order)
 
         df_scores_point.reset_index(inplace=True)
-        df_scores_point['Time order'] = df_scores_point['Time point'].map(time_order)
-        df_scores_point.sort_values(by=['Group', 'Time order'], inplace=True)
-
         err_df.reset_index(inplace=True)
-        err_df['Time order'] = err_df['Time point'].map(time_order)
+
+        df_scores_point.sort_values(by=['Group', 'Time order'], inplace=True)
         err_df.sort_values(by=['Group', 'Time order'], inplace=True)
 
+        # Default color palette if not provided
+        if color_dict is None:
+            import plotly.colors as plotly_color
+            palette = sum([getattr(plotly_color.qualitative, name) for name in ['Plotly', 'D3', 'Set2']], [])
+            groups = df_scores_['Group'].unique()
+            color_dict = {group: palette[i % len(palette)] for i, group in enumerate(groups)}
 
+        if symbol_dict is None:
+            symbol_dict = {}
 
-        #check color_dict must be a dictionary
-        if color_dict is not None:
-            if not isinstance(color_dict, dict):
-                raise ValueError('color_dict must be a dictionary')
-        else:
-            color_dict = None
+        r2 = self.df_explained_variance_
+        q2_test = self.q2_test
 
+        # Main line+error plot
+        fig = px.line(df_scores_point, x=pc[0], y=pc[1], line_group='Group',
+                    error_x=err_df[pc[0]], error_y=err_df[pc[1]],
+                    color='Group', color_discrete_map=color_dict,
+                    symbol='Time point', symbol_map=symbol_dict,
+                    title=f'<b>Principle component analysis ({self.scaling_method})<b>',
+                    height=fig_height, width=fig_width,
+                    labels={
+                        pc[0]: f"{pc[0]} R<sup>2</sup>X: {np.round(r2.loc[r2['PC'] == pc[0], 'Explained variance'].values[0]*100, 2)} %",
+                        pc[1]: f"{pc[1]} R<sup>2</sup>X: {np.round(r2.loc[r2['PC'] == pc[1], 'Explained variance'].values[0]*100, 2)} %",
+                        'Group': legend_name[0],
+                        'Time point': legend_name[1]
+                    })
 
-        if symbol_dict is not None:
-            if not isinstance(symbol_dict, dict):
-                raise ValueError('symbol_dict must be a dictionary')
-        else:
-            symbol_dict = None
-        
-
-
-        #If user not input color_dict then get unique of label and create color_dict
-        if color_dict is not None:
-            color_dict = color_dict
-        else:
-
-            import plotly.colors as plotly_colour
-            name_color_set = ['Plotly', 'D3', 'G10', 'T10', 'Alphabet', 'Dark24', 'Light24', 'Set1', 'Pastel1', 
-                                'Dark2', 'Set2', 'Pastel2', 'Set3', 'Antique', 'Safe', 'Bold', 'Pastel', 
-                                'Vivid', 'Prism']
-
-            palette = []
-            for name in name_color_set:
-                palette += getattr(plotly_colour.qualitative, name) # This is a list of colors
-
-
-            color_dict = {i: palette[i] for i in range(len(df_scores_point['Group'].unique()))}
-            
-
-        #new color_dict to match with unique label
-        group_unique = df_scores_point['Group'].unique()
-        #change key of color_dict to match with unique label
-        color_dict_2 = {group_unique[i]: list(color_dict.values())[i] for i in range(len(group_unique))}
-        
-
-        fig = px.line(df_scores_point, x=pc[0], y=pc[1], line_group='Time point', 
-                        error_x=err_df[pc[0]], error_y=err_df[pc[1]],
-                        color='Group', color_discrete_map=color_dict_2,
-                        symbol='Time point', symbol_map=symbol_dict,
-                        title='<b>Principle component analysis ({})<b>'.format(self.scaling_method), 
-                        height=fig_height, width=fig_width,
-                        labels={
-                            pc[0]: "{} R<sup>2</sup>X: {} %".format(pc[0], np.round(r2.loc[r2.loc[r2['PC']==pc[0]].index, 'Explained variance'].values[0]*100, decimals=2)),
-                            pc[1]: "{} R<sup>2</sup>X: {} %".format(pc[1], np.round(r2.loc[r2.loc[r2['PC']==pc[1]].index, 'Explained variance'].values[0]*100, decimals=2)),
-                            'Group': legend_name[0],
-                            'Time point': legend_name[1],
-                            })
-
-
-        
-
-
-
-        for connect_line in range(len(group_unique)):
-            # create a new trace for the connecting line
+        # Connect lines
+        for group in df_scores_point['Group'].unique():
+            df_group = df_scores_point[df_scores_point['Group'] == group]
             fig.add_trace(go.Scatter(
-                x=df_scores_point.loc[list(df_scores_point.loc[df_scores_point['Group'] == df_scores_point['Group'].unique()[connect_line]].index), pc[0]], # x-coordinates of the line
-                y=df_scores_point.loc[list(df_scores_point.loc[df_scores_point['Group'] == df_scores_point['Group'].unique()[connect_line]].index), pc[1]], # y-coordinates of the line
-                mode='lines', # specify the trace type as lines
-                line=dict(color=color_dict_2[group_unique[connect_line]], width=2), # set the color_ and width of the line
-                showlegend=False # hide the trace from the legend
+                x=df_group[pc[0]], y=df_group[pc[1]],
+                mode='lines',
+                line=dict(color=color_dict[group], width=2),
+                showlegend=False
             ))
 
-
-
-        fig.add_annotation(dict(font=dict(color="black",size=font_size),
-                                #x=x_loc,
-                                x=1.0,
-                                y=0.05,
-                                showarrow=False,
-                                text=f"<b>R<sup>2</sup>X (Cum): {np.round(r2.loc[r2.loc[r2['PC']==pc[1]].index, 'Cumulative variance'].values[0]*100, decimals=2)}%<b>",
-                                textangle=0,
-                                xref="paper",
-                                yref="paper"),
-                                # set alignment of text to left side of entry
-                                align="left")
-
-        fig.add_annotation(dict(font=dict(color="black",size=font_size),
-                                #x=x_loc,
-                                x=1.0,
-                                y=0.01,
-                                showarrow=False,
-                                text=f"<b>Q<sup>2</sup>X (Cum): {np.round(q2_test*100, decimals=2)}%<b>",
-                                textangle=0,
-                                xref="paper",
-                                yref="paper"),
-                                # set alignment of text to left side of entry
-                                align="left")
-
+        # Ellipse for global trajectory center
         fig.add_shape(type='path',
-                path=confidence_ellipse(df_scores_point[pc[0]], df_scores_point[pc[1]]))
+                    path=confidence_ellipse(df_scores_point[pc[0]], df_scores_point[pc[1]]),
+                    line=dict(color='black', width=2))
 
-
-                    #update axis as scitifics
-        fig.update_xaxes(tickformat=".1e")
-        fig.update_yaxes(tickformat=".1e")
-
-
-
-        fig.update_xaxes(zeroline=True, zerolinewidth=2, zerolinecolor='Black')
-        fig.update_yaxes(zeroline=True, zerolinewidth=2, zerolinecolor='Black')
-        fig.update_xaxes(showline=True, linewidth=2, linecolor='black')
-        fig.update_yaxes(showline=True, linewidth=2, linecolor='black')
+        # Axis and annotations
+        fig.update_xaxes(tickformat=".1e", zeroline=True, zerolinewidth=2, zerolinecolor='Black', showline=True, linewidth=2, linecolor='black')
+        fig.update_yaxes(tickformat=".1e", zeroline=True, zerolinewidth=2, zerolinecolor='Black', showline=True, linewidth=2, linecolor='black')
         fig.update_layout(
-            title={
-                'y':0.95,
-                'x':0.5,
-                'xanchor': 'center',
-                'yanchor': 'top'},
-            font=dict(size=title_font_size))
+            title={'y': 0.95, 'x': 0.5, 'xanchor': 'center', 'yanchor': 'top'},
+            font=dict(size=title_font_size),
+            paper_bgcolor='rgba(0,0,0,0)',
+            plot_bgcolor='rgba(0,0,0,0)'
+        )
 
+        fig.update_traces(marker=dict(size=marker_size, opacity=marker_opacity, line=dict(width=2, color='DarkSlateGrey')))
 
-        fig.update_traces(marker=dict(size=marker_size, 
-                                      opacity=marker_opacity, 
-                                      line=dict(width=2, 
-                                                color='DarkSlateGrey')))
+        fig.add_annotation(dict(font=dict(color="black", size=font_size),
+                                x=1.0, y=0.05, showarrow=False,
+                                text=f"<b>R<sup>2</sup>X (Cum): {np.round(r2.loc[r2['PC'] == pc[1], 'Cumulative variance'].values[0]*100, 2)}%<b>",
+                                xref="paper", yref="paper"))
+        fig.add_annotation(dict(font=dict(color="black", size=font_size),
+                                x=1.0, y=0.01, showarrow=False,
+                                text=f"<b>Q<sup>2</sup>X (Cum): {np.round(q2_test*100, 2)}%<b>",
+                                xref="paper", yref="paper"))
 
-        fig.update_layout(paper_bgcolor='rgba(0,0,0,0)',plot_bgcolor='rgba(0,0,0,0)')
-        
-        
         return fig
     
 
     def plot_3d_pca(self, pc=['PC1', 'PC2', 'PC3'], color_=None, color_dict=None, 
                     symbol_=None, symbol_dict=None, fig_height=900, fig_width=1300,
-                    marker_size=35, marker_opacity=0.7, font_size=20, title_font_size=20,
+                    marker_size=35, marker_opacity=0.7, marker_label=None, font_size=20, title_font_size=20,
                     legend_name=['Group', 'Time point']):
         import plotly.express as px
         '''
@@ -1947,10 +1949,11 @@ class pca:
 
         #If user not input color_dict then get unique of label and create color_dict
         if color_dict is not None:
-            color_dict = color_dict
+            color_dict_2 = color_dict
         else:
-
+            
             import plotly.colors as plotly_colour
+
             name_color_set = ['Plotly', 'D3', 'G10', 'T10', 'Alphabet', 'Dark24', 'Light24', 'Set1', 'Pastel1', 
                                 'Dark2', 'Set2', 'Pastel2', 'Set3', 'Antique', 'Safe', 'Bold', 'Pastel', 
                                 'Vivid', 'Prism']
@@ -1959,17 +1962,16 @@ class pca:
             for name in name_color_set:
                 palette += getattr(plotly_colour.qualitative, name) # This is a list of colors
 
-
             color_dict = {i: palette[i] for i in range(len(df_scores_['Group'].unique()))}
 
-        #new color_dict to match with unique label
-        group_unique = df_scores_['Group'].unique()
-        #change key of color_dict to match with unique label
-        color_dict_2 = {group_unique[i]: list(color_dict.values())[i] for i in range(len(group_unique))}
+            group_unique = df_scores_['Group'].unique()
+            color_dict_2 = {group_unique[i]: list(color_dict.values())[i] for i in range(len(group_unique))}
+        
 
 
         fig = px.scatter_3d(df_scores_, x=pc[0], y=pc[1], z=pc[2], color='Group', symbol=symbol_, 
                             color_discrete_map=color_dict_2, symbol_map=symbol_dict, 
+                            text=marker_label,
                             title=f'<b>PCA Scores Plot<b> {self.scaling_method} scaling', 
                             height=fig_height, width=fig_width,
                             labels={'color': legend_name[0], 'symbol': legend_name[1],
@@ -1983,6 +1985,9 @@ class pca:
                             opacity=marker_opacity,
                             line=dict(width=2, color='DarkSlateGrey')))
 
+        fig.update_traces(textposition='middle center',
+                    textfont_size=marker_size-(0.4*marker_size))
+
         fig.add_annotation(dict(font=dict(color="black",size=font_size), x=1.0, y=0.05, showarrow=False,
                                 text=f"<b>R<sup>2</sup>X (Cum): {np.round(r2.loc[r2.loc[r2['PC']==pc[2]].index, 'Cumulative variance'].values[0]*100, decimals=2)}%<b>",
                                 textangle=0, xref="paper", yref="paper"), align="left")
